/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.io.StringReader;
import java.util.List;

/**
 * Represents a chromosome that is one sequence in a genome (i.e. one sequence
 * in a FASTA file).
 */
public class Chromosome implements Comparable<Chromosome> {

  // Reference to the parent genome of this chromosome
  private Genome genome;

  // The name of this chromosome
  private String name;

  // The length of this chromosome
  private long   length;

  // The offset to the start of this chromosome in the parent genome's FASTA file
  private long   offset;

  // The line length of this chromosome in the parent genome's FASTA file
  private int    lineLength;

  // The bytes per line for this chromosome in the parent genome's FASTA file
  private int    lineBytes;

  /**
   * Creates a new Chromosome object.
   *
   * @param genome
   *          A reference to the parent genome of this chromosome
   * @param name
   *          The name of this chromsome
   * @param length
   *          The length of this chromosome
   * @param offset
   *          The offset to the start of this chromosome in the parent genome's
   *          FASTA file
   * @param lineLength
   *          The line length of this chromosome in the parent genome's FASTA
   *          file
   * @param lineBytes
   *          The bytes per line for this chromosome in the parent genome's
   *          FASTA file
   */
  public Chromosome(Genome genome, String name, long length, long offset, int lineLength, int lineBytes) {

    // Throw errors for null objects
    if (genome == null) {
      throw new IllegalArgumentException("Null genome specified");
    }
    if (name == null) {
      throw new IllegalArgumentException("Null chromosome name specified");
    }

    // Initialize variables
    this.genome = genome;
    this.name = name;
    this.length = length;
    this.offset = offset;
    this.lineLength = lineLength;
    this.lineBytes = lineBytes;

  }

  /**
   * Returns a reference to the parent genome of this chromosome.
   *
   * @return A Genome object representing the parent genome of this chromosome
   */
  public Genome getGenome() {
    return genome;
  }

  /**
   * Returns the name of this chromosome.
   *
   * @return the name of this chromosome
   */
  public String getName() {
    return name;
  }

  /**
   * Returns the length of this chromosome.
   *
   * @return the length of this chromosome
   */
  public long getLength() {
    return length;
  }

  /**
   * Returns the sequence string starting at startLoc and ending at endLoc (both
   * inclusive) for this chromosome. Can only work for sequences roughly the
   * size of 2^31 (~2,147,000,000) base pairs long or shorter.
   *
   * @param startLoc
   *          The starting location of the desired sequence (inclusive)
   * @param endLoc
   *          The ending location of the desired sequence (inclusive)
   * @return The sequence string starting at startLoc and ending at endLoc (both
   *         inclusive) for this chromosome
   * @throws IOException
   *           If there are problems reading from the FASTA file
   */
  public String getSequence(long startLoc, long endLoc) throws IOException {

    // Check for valid startLoc and endLoc parameters
    if (endLoc > length || startLoc < 1 || startLoc > endLoc) {
      throw new IllegalArgumentException("Improper start and/or end locations specified for chromosome '" + name + "':"
          + startLoc + "-" + endLoc);
    }

    // The byte containing the first base in the desired sequence string
    // If startLoc is at the end of a line, just use the offset + num of full lines * lineBytes - newline bytes
    // If startLoc is before the end of a line, use the offset + num of full lines * lineBytes + leftover chars on last line
    long startByte;
    long numFullLinesBeforeStart = startLoc / lineLength;
    if (startLoc % lineLength == 0) {
      startByte = offset + numFullLinesBeforeStart * lineBytes - (lineBytes - lineLength);
    } else {
      startByte = offset + numFullLinesBeforeStart * lineBytes + startLoc % lineLength;
    }

    // The byte containing the last base in the desired sequence string
    // If endLoc is at the end of a line, just use the offset + num of full lines * lineBytes - newline bytes
    // If endLoc is before the end of a line, use the offset + num of full lines * lineBytes + leftover chars on last line
    long endByte;
    long numFullLinesBeforeEnd = endLoc / lineLength;
    if (endLoc % lineLength == 0) {
      endByte = offset + numFullLinesBeforeEnd * lineBytes - (lineBytes - lineLength);
    } else {
      endByte = offset + numFullLinesBeforeEnd * lineBytes + endLoc % lineLength;
    }

    // The number of bytes to read from the FASTA file
    long requiredBytes = endByte - startByte + 1;

    // Check to make sure the desired number of bytes will fit in a StringBuffer
    if (requiredBytes > Integer.MAX_VALUE) {
      throw new IllegalArgumentException("Can't retrieve a sequence as big as " + name + ":" + startLoc + "-" + endLoc);
    }

    // Initialize empty StringBuffer
    StringBuffer seq = new StringBuffer((int) requiredBytes);

    try {

      // Get RandomAccessFile object and skip to startByte
      RandomAccessFile faFileReader = genome.getFaFileReader();
      faFileReader.seek(startByte - 1);

      // Read requiredBytes number of bytes
      byte[] bytes = new byte[(int) requiredBytes];
      faFileReader.readFully(bytes);

      // Set up BufferedReader to process array of bytes
      // BufferedReader has a larger buffer than Scanner and seemed to be faster
      BufferedReader reader = new BufferedReader(new StringReader(new String(bytes)));

      // Append all lines in the BufferedReader to the StringBuffer
      String line = reader.readLine();
      while (line != null) {
        seq.append(line);
        line = reader.readLine();
      }

      // Close BufferedReader object
      reader.close();

      // Return desired sequence string
      return seq.toString();

      // Handle any IO errors by throwing a generic IOException message
    } catch (IOException e) {
      throw new IOException("Problem reading from the FASTA file", e);
    }

  }

  /**
   * Returns the full sequence string for this chromosome. Can only work for
   * chromosomes roughly the size of 2^31 (~2,147,000,000) base pairs long or
   * shorter.
   *
   * @return The full sequence string for this chromosome
   * @throws IOException
   *           If there are problems reading from the FASTA file
   */
  public String getFullSequence() throws IOException {
    return getSequence(1, length);
  }

  @Override
  public int compareTo(Chromosome other) {

    // Return 0 if chromosomes are equal
    if (equals(other)) {
      return 0;
    }

    // Ensure both chromosomes have the same Genome object
    if (genome != other.getGenome()) {
      throw new IllegalArgumentException("Illegal comparison between chromosomes from different genomes");
    }

    // Find indices of chromosomes in the chromosome list
    int thisIdx = -1;
    int otherIdx = -1;
    List<Chromosome> chrList = genome.getChromosomeList();
    for (int i = 0; i < chrList.size(); i++) {
      if (chrList.get(i) == this) {
        if (thisIdx == -1) {
          thisIdx = i;
        } else {
          throw new RuntimeException("Unexpected behavior: Chromosome '" + getName()
              + "' exists multiple times in the genome's chromosome list");
        }
      }
      if (chrList.get(i) == other) {
        if (otherIdx == -1) {
          otherIdx = i;
        } else {
          throw new RuntimeException("Unexpected behavior: Chromosome '" + other.getName()
              + "' exists multiple times in the genome's chromosome list");
        }
      }
    }

    // Ensure that indices are different and that both chromosomes are found in the chromosome list
    if (thisIdx == otherIdx || thisIdx == -1 || otherIdx == -1) {
      throw new RuntimeException(
          "Unexpected behavior while searching for chromosome indices in the genome's chromosome list");
    }

    // Return -1 if this chromosome occurs first in the chromosome list, +1 otherwise
    if (thisIdx < otherIdx) {
      return -1;
    } else {
      return 1;
    }

  }

  @Override
  public boolean equals(Object obj) {
    if (obj == null || !(obj instanceof Chromosome)) {
      return false;
    } else {
      Chromosome other = (Chromosome) obj;
      return genome == other.genome && name.equals(other.name) && length == other.length && offset == other.offset
          && lineLength == other.lineLength && lineBytes == other.lineBytes;
    }
  }

}
