/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio.shear;

import java.util.Arrays;

/**
 * This is the main SHEAR class. It is used to read the initial command line
 * arguments, launch the appropriate command module, and handle errors.
 */
public class Shear {

  public static final String VERSION = "1.0.0";

  /**
   * Display the usage syntax for SHEAR.
   */
  public static void mainHelp() {
    // Maximum length of println line should be 107 (4 indent + 20 before text + 3 after text + 80 text)
    // This will display 108 when clicked at the end
    System.out.println();
    System.out.println("SHEAR (Sample Heterogeneity Estimation and Assembly by Reference)");
    System.out.println("Version:  " + Shear.VERSION);
    System.out.println("Website:  http://vk.cs.umn.edu/SHEAR/");
    System.out.println();
    System.out.println("Usage:");
    System.out.println("    java -jar SHEAR.jar <command> [options]");
    System.out.println();
    System.out.println("commands:");
    System.out.println("    sv");
    System.out.println("    assemble");
    System.out.println();
    System.out.println("sv options:");
    System.out.println("    -p <prefix>                        Prefix for all generated files.");
    System.out.println("                                       REQUIRED");
    System.out.println();
    System.out.println("    -b <bam_file>                      BAM alignment file to be used. Must have");
    System.out.println("                                       BAM index file (i.e. *.bam.bai) in the");
    System.out.println("                                       same directory.");
    System.out.println("                                       REQUIRED");
    System.out.println();
    System.out.println("    -f <fa_reference_file>             Reference file (FASTA format) to be used.");
    System.out.println("                                       Must have index file (i.e. *.fa.fai) and");
    System.out.println("                                       dictionary file (i.e. *.dict) in the same");
    System.out.println("                                       directory.");
    System.out.println("                                       REQUIRED");
    System.out.println();
    System.out.println("    --sv-algo [crest|delly]            SV prediction alorithm to use. This can");
    System.out.println("                                       be set to 'crest' or 'delly'. If 'crest'");
    System.out.println("                                       is used, then the -t parameter is");
    System.out.println("                                       required (see below).");
    System.out.println("                                       Optional (default: 'delly')");
    System.out.println();
    System.out.println("    --align-algo [bwasw|bwamem]        Alignment algorithm to use for local");
    System.out.println("                                       realigment. This can be set to 'bwasw' or");
    System.out.println("                                       'bwamem'");
    System.out.println("                                       Optional (default: 'bwamem')");
    System.out.println();
    System.out.println("    -t <two_bit_reference_file>        Reference file (2bit format) to be used.");
    System.out.println("                                       Required if --sv-algo is set to 'crest'");
    System.out.println();
    System.out.println("    -i <bwa_index_prefix>              Prefix of the BWA index files for this");
    System.out.println("                                       reference (i.e. <bwa_index_prefix>.sa,");
    System.out.println("                                       <bwa_index_prefix>.pac, etc.).");
    System.out.println("                                       Optional (default: <fa_reference_file>)");
    System.out.println();
    System.out.println("    -r <region>                        Region of the input alignment to analyze");
    System.out.println("                                       (i.e. \"chr1\" for full chromosome or");
    System.out.println("                                       \"chr1:10000-15000\" for a particular");
    System.out.println("                                       region, including endpoints). Coordinates");
    System.out.println("                                       are 1-based.");
    System.out.println("                                       Optional (default: whole genome)");
    System.out.println();
    System.out.println("    --preds <pred_file>                Pre-existing SV prediction file generated");
    System.out.println("                                       for this alignment. The format of this");
    System.out.println("                                       file should match whichever SV prediction");
    System.out.println("                                       algorithm is selected (i.e. *.predSV.txt");
    System.out.println("                                       for CREST, or *.vcf for DELLY). Can be");
    System.out.println("                                       useful to use if the SV prediction");
    System.out.println("                                       algorithm has already been run on this");
    System.out.println("                                       alignment independently, or if SHEAR was");
    System.out.println("                                       interrupted and you want to continue from");
    System.out.println("                                       the previous intermediary results. Note");
    System.out.println("                                       that DELLY's multiple VCF files are");
    System.out.println("                                       combined when run from SHEAR, and that is");
    System.out.println("                                       the expected format if specified here.");
    System.out.println("                                       Optional (default: none)");
    System.out.println();
    System.out.println("    --min-het <min_het_threshold>      Minimum estimated heterogeneity level for");
    System.out.println("                                       a variant to be included in the results.");
    System.out.println("                                       Must be a numeric value between 0 and 1");
    System.out.println("                                       (i.e. \"--min-het 0.2\" would only output");
    System.out.println("                                       variants that have an estimated");
    System.out.println("                                       heterogeneity level of at least 20%).");
    System.out.println("                                       Optional (default: none)");
    System.out.println();
    System.out.println("    --preprocess                       Preprocesses the input BAM file by fixing");
    System.out.println("                                       potential errors, sorting, and ensuring");
    System.out.println("                                       read group information is present. May be");
    System.out.println("                                       necessary if these tasks have not already");
    System.out.println("                                       been done rpeviously.");
    System.out.println("                                       Optional (default: OFF)");
    System.out.println();
    System.out.println("    --sv-only                          \"SV Only\" prediction mode. Will skip");
    System.out.println("                                       SNP/INDEL prediction and only return");
    System.out.println("                                       *.sdi and *.report results for SVs.");
    System.out.println("                                       Optional (default: OFF)");
    System.out.println();
    System.out.println("    -d                                 Debugging mode. Will keep additional");
    System.out.println("                                       intermediary files and produce more");
    System.out.println("                                       verbose output information.");
    System.out.println("                                       Optional (default: OFF)");
    System.out.println();
    System.out.println("assemble options:");
    System.out.println("    -s <sdi_file>                      SDI file produced by SHEAR's 'sv' command");
    System.out.println("                                       containing the SVs to use to create the");
    System.out.println("                                       new genomic sequence.");
    System.out.println("                                       REQUIRED");
    System.out.println();
    System.out.println("    -f <fasta_reference_file>          Reference file (FASTA format) to be used");
    System.out.println("                                       for guiding assembly. Must have index");
    System.out.println("                                       file (i.e. *.fa.fai) in the same");
    System.out.println("                                       directory.");
    System.out.println("                                       REQUIRED");
    System.out.println();
    System.out.println("    -o <output_fasta_file>             FASTA file to output.");
    System.out.println("                                       REQUIRED");
    System.out.println();
    System.out.println("    -d                                 Debugging mode. Will keep additional");
    System.out.println("                                       intermediary files and produce more");
    System.out.println("                                       verbose output information.");
    System.out.println("                                       Optional (default: OFF)");
    System.out.println();
  }

  /**
   * Display a usage error message and exit. Used when the user does not specify
   * proper arguments.
   *
   * @param message
   *          The message to display to the user regarding what went wrong
   */
  public static void usageError(String message) {
    log("!----- ERROR -----!", false);
    log(message, false);
    log("SHEAR usage syntax:", false);
    mainHelp();
    System.exit(1);
  }

  /**
   * Display a runtime error message and exit. Used when there is some
   * unexpected problem with SHEAR.
   *
   * @param e
   *          Exception that triggered the runtime error
   */
  public static void runtimeError(Exception e) {
    log("!----- ERROR -----!", false);
    log("SHEAR has encountered an unexpected problem", false);
    log("Please contact the developers if the problem persists", false);
    e.printStackTrace();
    log("Exiting...", false);
    System.exit(1);
  }

  public static void log(String message, boolean skipNextLine) {
    System.out.println("[SHEAR] " + message);
    if (skipNextLine) {
      System.out.println("[SHEAR]");
    }
  }

  /**
   * Main method to check for command, pass the rest of the arguments to the
   * proper command module, and handle error catching.
   *
   * @param args
   *          Command line arguments passed to SHEAR
   */
  public static void main(String[] args) {

    try {

      // Check for existence of command
      if (args.length == 0) {
        throw new ShearUsageException("Must specify a command");

        // Run SHEAR-SV
      } else if (args[0].equals("sv")) {
        ShearSV shearSV = new ShearSV(Arrays.copyOfRange(args, 1, args.length));
        shearSV.execute();

        // Run SHEAR-Assemble
      } else if (args[0].equals("assemble")) {
        ShearAssemble shearAssemble = new ShearAssemble(Arrays.copyOfRange(args, 1, args.length));
        shearAssemble.execute();

        // Unrecognized command
      } else {
        throw new ShearUsageException("Unrecognized command: " + args[0]);
      }

      // Catch SHEAR runtime exception
    } catch (ShearRuntimeException e) {
      Shear.runtimeError(e);

      // Catch SHEAR usage exception
    } catch (ShearUsageException e) {
      Shear.usageError(e.getMessage());

      // Catch other exceptions (shouldn't happen, so display a message about it)
    } catch (Exception e) {
      log("WARNING: Exception caught that is not ShearRuntimeException", false);
      Shear.runtimeError(e);
    }

  }

}
