/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio.shear;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Scanner;

/**
 * This is used to run SHEAR-Assemble. The constructor creates a new
 * ShearAssemble instance and parses all of the provided command-line arguments.
 * The program is run using the execute method.
 */
public class ShearAssemble {

  // Path to SDI file to use
  private String  sdiFile;

  // Path to FASTA reference file to use
  private String  origFaRefFile;

  // Path to new FASTA reference file to output
  private String  newFaRefFile;

  // Whether or not debugging mode is activated
  private boolean debug;

  /**
   * Creates a new ShearAssemble object by initializing the local variables that
   * are read from the command line arguments.
   *
   * @param args
   *          The command line arguments to process, excluding the first one
   *          (SHEAR command, i.e. 'sv' or 'assemble')
   */
  public ShearAssemble(String[] args) throws ShearUsageException {

    // Initialize debug value
    debug = false;

    // Read in command-line arguments
    for (int i = 0; i < args.length; i++) {
      try {
        if (args[i].equals("-s")) {
          sdiFile = args[++i];
        } else if (args[i].equals("-f")) {
          origFaRefFile = args[++i];
        } else if (args[i].equals("-o")) {
          newFaRefFile = args[++i];
        } else if (args[i].equals("-d")) {
          debug = true;
        } else {
          throw new ShearUsageException("Unknown argument: " + args[i]);
        }
      } catch (ArrayIndexOutOfBoundsException e) {
        throw new ShearUsageException("Missing argument value for flag '" + args[i - 1] + "'");
      }
    }

    try {

      // Validate sdiFile
      if (sdiFile == null) {
        throw new ShearUsageException("Must provide an SDI file with the '-s' argument");
      }
      if (!new File(sdiFile).exists()) {
        throw new ShearUsageException("SDI file does not exist: " + sdiFile);
      } else {
        sdiFile = new File(sdiFile).getCanonicalPath();
      }

      // Validate origFaRefFile
      if (origFaRefFile == null) {
        throw new ShearUsageException("Must provide a FASTA reference file with the '-f' argument");
      }
      if (!new File(origFaRefFile).exists()) {
        throw new ShearUsageException("FASTA reference file does not exist: " + origFaRefFile);
      } else if (!new File(origFaRefFile + ".fai").exists()) {
        throw new ShearUsageException("FASTA reference index file does not exist: " + origFaRefFile + ".fai");
      } else {
        origFaRefFile = new File(origFaRefFile).getCanonicalPath();
      }

      // Validate newFaRefFile
      if (newFaRefFile == null) {
        throw new ShearUsageException("Must provide an output FASTA file with the '-o' argument");
      }
      newFaRefFile = new File(newFaRefFile).getCanonicalPath();

      // Handle errors
    } catch (IOException e) {
      throw new ShearRuntimeException("Problem occurred setting up SHEAR instance", e);
    }

    // Output settings information
    Shear.log("Starting SHEAR-Assemble (version " + Shear.VERSION + ")", false);
    Shear.log("SDI file:                " + sdiFile, false);
    Shear.log("FASTA reference file:    " + origFaRefFile, false);
    Shear.log("FASTA output file:       " + newFaRefFile, false);
    if (debug) {
      Shear.log("Debugging mode:          ON", false);
    } else {
      Shear.log("Debugging mode:          OFF", false);
    }

  }

  /**
   * Parses the next line from the specified SDI file input stream (provided as
   * a Scanner object) and returns an SdiLine object.
   *
   * @param sc
   *          The Scanner object that contains the SDI file input stream
   * @return The SdiLine object corresponding to the next line of the SDI file
   *         input
   */
  private SdiLine getNextMutation(Scanner sc) {
    if (sc.hasNextLine()) {
      String[] tokens = sc.nextLine().split("\t");
      String region = tokens[0];
      int loc = Integer.valueOf(tokens[1]);
      int diff = Integer.valueOf(tokens[2]);
      String origBases = tokens[3].equals("-") ? "" : tokens[3];
      String newBases = tokens[4].equals("-") ? "" : tokens[4];
      if (newBases.length() - origBases.length() != diff) {
        throw new ShearRuntimeException("Unexpected format in SDI file:" + sdiFile);
      }
      return new SdiLine(region, loc, origBases, newBases);
    } else {
      return null;
    }
  }

  /**
   * Returns true if the specified mutation occurs on the specified region and
   * has a location equal to or less than the number of bases that have been
   * processed for the region.
   *
   * @param nextMutation
   *          The mutation to check for (as an SdiLine object)
   * @param currentRegion
   *          The name of the current region being processed
   * @param regionBasesRead
   *          The number of bases that have been read from the current region
   * @return True if there is a mutation present to process, false otherwise
   */
  private boolean mutationPresent(SdiLine nextMutation, String currentRegion, int regionBasesRead) {
    return nextMutation != null && nextMutation.region.equals(currentRegion) && nextMutation.loc <= regionBasesRead;
  }

  /**
   * SdiLine is a helper class to store the information that appears on one line
   * of an SDI file.
   */
  private class SdiLine {

    // The region of the variant
    private String region;

    // The location of the variant
    private int    loc;

    // The original bases at the variant location (can be an empty string)
    private String origBases;

    // The new bases at the variant location (can be an empty string)
    private String newBases;

    /**
     * Creates a new SdiLine object.
     *
     * @param region
     *          The region of the variant
     * @param loc
     *          The location of the variant
     * @param origBases
     *          The original bases at the variant location (can be an empty
     *          string)
     * @param newBases
     *          The new bases at the variant location (can be an empty string)
     */
    private SdiLine(String region, int loc, String origBases, String newBases) {
      this.region = region;
      this.loc = loc;
      this.origBases = origBases;
      this.newBases = newBases;
    }

  }

  /**
   * Main method to run the ShearAssemble module.
   */
  public void execute() {

    try {

      // Set up I/O streams
      BufferedReader origFaRefIn = new BufferedReader(new FileReader(origFaRefFile));
      Scanner sdiIn = new Scanner(new File(sdiFile));
      FileWriter newFastaOut = new FileWriter(newFaRefFile);

      // Initialize variables
      String currentRegion = null;
      int regionBasesRead = 0;
      int processedAndAssembledBases = 0;
      StringBuffer toWrite = new StringBuffer(100000);
      SdiLine nextMutation = getNextMutation(sdiIn);

      // Loop through each line in original FASTA file
      String line = origFaRefIn.readLine();
      while (line != null) {

        // Check for new region (i.e. FASTA sequence header line)
        if (line.startsWith(">")) {

          // Write anything remaining to write
          if (toWrite.length() > 0) {
            newFastaOut.write(toWrite + "\n");
            toWrite.delete(0, toWrite.length());
          }

          // Get the current region name (only part before any whitespace)
          Scanner sequenceHeaderLineScanner = new Scanner(line);
          currentRegion = sequenceHeaderLineScanner.next().substring(1);
          sequenceHeaderLineScanner.close();

          // Reset regionBasesRead and processedAndAssembledBases
          regionBasesRead = 0;
          processedAndAssembledBases = 0;

          // Log message
          Shear.log("[SHEAR] Parsing region '" + currentRegion + "'", debug);

          // Write header line (full line including whitespace)
          newFastaOut.write(line + "\n");

          // Process regular (non-header) line
        } else {

          // Add bases to toWrite queue
          toWrite.append(line);
          regionBasesRead += line.length();

          // Check for any relevant mutations in the bases in the toWrite queue
          while (mutationPresent(nextMutation, currentRegion, regionBasesRead)) {

            // Basepair difference for mutation
            int diff = nextMutation.newBases.length() - nextMutation.origBases.length();

            // Log message
            Shear.log(
                "  Processing mutation: " + nextMutation.region + ":" + nextMutation.loc + "\tbp change: " + diff,
                debug);

            // Skip mutation if it overlaps with a previously assembled mutation
            if (nextMutation.loc <= processedAndAssembledBases) {
              Shear.log("    Skipping this mutation. It overlaps with a previously processed mutation.", debug);

              // Mutation does not overlap with anything previously processed
            } else {

              // Read bases into toWrite queue until everything in nextMutations's original bases are in queue
              while (regionBasesRead < nextMutation.loc + nextMutation.origBases.length() - 1) {
                line = origFaRefIn.readLine();
                toWrite.append(line);
                regionBasesRead += line.length();
              }

              // Indices in toWrite queue to start and end replacement
              int i1 = toWrite.length() - 1 - (regionBasesRead - nextMutation.loc);
              int i2 = toWrite.length() - 1 - (regionBasesRead - nextMutation.loc) + nextMutation.origBases.length();

              // Validate that the bases in the toWrite queue at these indices match the original bases of the mutation
              if (!toWrite.substring(i1, i2).equals(nextMutation.origBases)) {
                origFaRefIn.close();
                sdiIn.close();
                newFastaOut.close();
                throw new ShearRuntimeException("SDI file does not match reference sequence in file " + sdiFile);
              }

              // Replace original bases with new bases for this mutation in the toWrite queue
              toWrite.replace(i1, i2, nextMutation.newBases);

              // Update processedAndAssembledBases
              processedAndAssembledBases = nextMutation.loc + nextMutation.origBases.length() - 1;

            }

            // Get the next mutation from the SDI input
            nextMutation = getNextMutation(sdiIn);

          }

          // Write any lines to output
          while (toWrite.length() >= 60) {
            newFastaOut.write(toWrite.substring(0, 60) + "\n");
            toWrite.delete(0, 60);
          }

        }

        // Grab the next line in original FASTA file
        line = origFaRefIn.readLine();

      }

      // Write anything remaining to write
      if (toWrite.length() > 0) {
        newFastaOut.write(toWrite + "\n");
        toWrite.delete(0, toWrite.length());
      }

      // Close I/O streams
      Shear.log("All Finished!", false);
      origFaRefIn.close();
      sdiIn.close();
      newFastaOut.close();

      // Catch errors
    } catch (Exception e) {
      throw new ShearRuntimeException("Error while running SHEAR-Assemble", e);

    }

  }

}
