/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio.shear.heterogeneity;

/**
 * BreakpointReads represents the respective number of reads that are clipped at
 * a particular breakpoint (in either direction) and that span the breakpoint.
 */
public class BreakpointReads {

  // The number of reads that are clipped in the 5' direction of this breakpoint (i.e. clipped on left side of breakpoint)
  private int fivePrimeClippedReads;

  // The number of reads that are clipped in the 3' direction of this breakpoint (i.e. clipped on right side of breakpoint)
  private int threePrimeClippedReads;

  // The number of reads that span this breakpoint
  private int spanningReads;

  /**
   * Creates a new BreakpointReads object with the specified number of clipped
   * and spanning reads.
   *
   * @param fivePrimeClippedReads
   *          The number of reads that are clipped in the 5' direction of this
   *          breakpoint (i.e. clipped on left side of breakpoint)
   * @param fivePrimeClippedReads
   *          The number of reads that are clipped in the 3' direction of this
   *          breakpoint (i.e. clipped on right side of breakpoint)
   * @param spanningReads
   *          The number of reads that span this breakpoint
   */
  public BreakpointReads(int fivePrimeClippedReads, int threePrimeClippedReads, int spanningReads) {
    this.fivePrimeClippedReads = fivePrimeClippedReads;
    this.threePrimeClippedReads = threePrimeClippedReads;
    this.spanningReads = spanningReads;
  }

  /**
   * Returns the number of reads that are clipped in the 5' direction of this
   * breakpoint (i.e. clipped on left side of breakpoint).
   *
   * @return The number of reads that are clipped in the 5' direction of this
   *         breakpoint (i.e. clipped on left side of breakpoint)
   */
  public int getFivePrimeClippedReads() {
    return fivePrimeClippedReads;
  }

  /**
   * Returns the number of reads that are clipped in the 3' direction of this
   * breakpoint (i.e. clipped on right side of breakpoint).
   *
   * @return The number of reads that are clipped in the 3' direction of this
   *         breakpoint (i.e. clipped on right side of breakpoint)
   */
  public int getThreePrimeClippedReads() {
    return threePrimeClippedReads;
  }

  /**
   * Returns the number of reads that span this breakpoint.
   *
   * @return The number of reads that span this breakpoint
   */
  public int getSpanningReads() {
    return spanningReads;
  }

  /**
   * Increments the number of reads that are clipped in the 5' direction of this
   * breakpoint by the specified amount.
   *
   * @param toAdd
   *          The amount to increment the 5' clipped reads by
   */
  public void addToFivePrimeClippedReads(int toAdd) {
    fivePrimeClippedReads += toAdd;
  }

  /**
   * Increments the number of reads that are clipped in the 3' direction of this
   * breakpoint by the specified amount.
   *
   * @param toAdd
   *          The amount to increment the 3' clipped reads by
   */
  public void addToThreePrimeClippedReads(int toAdd) {
    threePrimeClippedReads += toAdd;
  }

  /**
   * Increments the number of reads that span this breakpoint by the specified
   * amount.
   *
   * @param toAdd
   *          The amount to increment the spanning reads by
   */
  public void addToSpanningReads(int toAdd) {
    spanningReads += toAdd;
  }

  @Override
  public String toString() {
    return String.format("%d/%d/%d", fivePrimeClippedReads, threePrimeClippedReads, spanningReads);
  }

}
