/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio.shear.heterogeneity;

import htsjdk.samtools.CigarElement;
import htsjdk.samtools.CigarOperator;

import java.io.PrintStream;

import org.broadinstitute.gatk.engine.CommandLineGATK;
import org.broadinstitute.gatk.engine.walkers.DataSource;
import org.broadinstitute.gatk.engine.walkers.ReadWalker;
import org.broadinstitute.gatk.engine.walkers.Requires;
import org.broadinstitute.gatk.utils.commandline.Argument;
import org.broadinstitute.gatk.utils.commandline.Output;
import org.broadinstitute.gatk.utils.contexts.ReferenceContext;
import org.broadinstitute.gatk.utils.help.DocumentedGATKFeature;
import org.broadinstitute.gatk.utils.refdata.RefMetaDataTracker;
import org.broadinstitute.gatk.utils.sam.GATKSAMRecord;

/**
 * BreakpointReadsWalker is an extension of GATK's ReadWalker that will count
 * the number of reads that span a given locus versus those that are clipped in
 * a either direction at the given locus. This is used by the getBreakpointReads
 * method in the SvHeterogeneityEstimator class.
 */
@DocumentedGATKFeature(groupName = "Quality Control and Simple Analysis Tools", extraDocs = { CommandLineGATK.class })
@Requires({ DataSource.READS, DataSource.REFERENCE })
public class BreakpointReadsWalker extends ReadWalker<BreakpointReads, BreakpointReads> {

  @Output
  protected PrintStream out;

  @Argument(
      fullName = "clipLocus",
      shortName = "clipLocus",
      doc = "Locus of the breakpoint (left side).",
      required = true)
  protected int         LOCUS                   = -1;

  @Argument(
      fullName = "unclippedSpanningOnly",
      shortName = "unclippedSpanningOnly",
      doc = "Only count spanning reads that are completely unclipped (i.e. not just at the breakpoint).",
      required = false)
  protected boolean     UNCLIPPED_SPANNING_ONLY = false;

  @Override
  public BreakpointReads reduceInit() {
    return new BreakpointReads(0, 0, 0);
  }

  @Override
  public BreakpointReads reduce(BreakpointReads value, BreakpointReads sum) {
    sum.addToFivePrimeClippedReads(value.getFivePrimeClippedReads());
    sum.addToThreePrimeClippedReads(value.getThreePrimeClippedReads());
    sum.addToSpanningReads(value.getSpanningReads());
    return sum;
  }

  @Override
  public BreakpointReads map(ReferenceContext ref, GATKSAMRecord read, RefMetaDataTracker tracker) {

    // If the read is unmapped, then return 0/0/0
    if (read.getReadUnmappedFlag()) {
      return new BreakpointReads(0, 0, 0);
    }

    // If the first CIGAR element is soft-clipping, and the alignment starts right after breakpoint, then return 0/1/0
    if (read.getAlignmentStart() == LOCUS + 1 && read.getCigar().getCigarElement(0).getOperator() == CigarOperator.S) {
      return new BreakpointReads(1, 0, 0);
    }

    // If the alignment starts after the breakpoint, then return 0/0/0
    if (read.getAlignmentStart() >= LOCUS + 1) {
      return new BreakpointReads(0, 0, 0);
    }

    // Initialize pos to alignment start position
    int pos = read.getAlignmentStart();

    // Boolean to keep track of if CIGAR has any clipping
    boolean isClipped = false;

    // Loop through CIGAR elements, updating isClipped boolean and the pos value
    for (CigarElement elt : read.getCigar().getCigarElements()) {
      switch (elt.getOperator()) {
        case H: // hard-clip: ignore
          isClipped = true;
          break;
        case S: // soft clip: return 0/1/0 if we are at the breakpoint
          isClipped = true;
          if (pos == LOCUS + 1) {
            return new BreakpointReads(0, 1, 0);
          }
          break;
        case EQ: // sequence match: pass to alignment match
        case X: // sequence mismatch: pass to alignment match
        case M: // alignment match: add to pos
          pos += elt.getLength();
          break;
        case N: // reference skip: pass to deletion
        case D: // deletion: add to pos
          pos += elt.getLength();
          break;
        case P: // padding: don't add to pos
          break;
        case I: // insertion: don't add to pos
          break;
        default:
          throw new IllegalStateException("Case statement didn't deal with cigar op: " + elt.getOperator());
      }
    }

    // If alignment didn't span the clip locus, then return 0/0/0
    if (pos == LOCUS + 1) {
      return new BreakpointReads(0, 0, 0);

      // If alignment spans clip locus but is clipped elsewhere and unclippedSpanningOnly is set, then return 0/0/0
    } else if (UNCLIPPED_SPANNING_ONLY && isClipped) {
      return new BreakpointReads(0, 0, 0);

      // If alignment spans clip locus and is unclipped everywhere or unclippedSpanningOnly is not set, then return 0/0/1
    } else {
      return new BreakpointReads(0, 0, 1);
    }

  }

  @Override
  public void onTraversalDone(BreakpointReads result) {

    // Output results to output file
    out.println(result.getFivePrimeClippedReads() + "\t" + result.getThreePrimeClippedReads() + "\t"
        + result.getSpanningReads());

    // Output results to standard output
    System.out.println(String.format(
        "Number of 5' end clipped reads at the breakpoint: %d",
        result.getFivePrimeClippedReads()));
    System.out.println(String.format(
        "Number of 3' end clipped reads at the breakpoint: %d",
        result.getThreePrimeClippedReads()));
    System.out.println(String.format("Number of spanning reads at the breakpoint:    %d", result.getSpanningReads()));

  }

}
