/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio.shear.heterogeneity;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.ArrayList;
import java.util.List;
import java.util.Scanner;
import java.util.Set;
import java.util.TreeSet;

import org.broadinstitute.gatk.utils.progressmeter.ProgressMeterDaemon;

import edu.umn.cs.kumarbio.shear.Shear;
import edu.umn.cs.kumarbio.shear.ShearRuntimeException;
import edu.umn.cs.kumarbio.shear.ShearSV;
import edu.umn.cs.kumarbio.shear.system.RunGatkCommand;
import edu.umn.cs.kumarbio.shear.variants.Deletion;
import edu.umn.cs.kumarbio.shear.variants.Insertion;
import edu.umn.cs.kumarbio.shear.variants.Inversion;
import edu.umn.cs.kumarbio.shear.variants.Location;
import edu.umn.cs.kumarbio.shear.variants.TandemDuplication;
import edu.umn.cs.kumarbio.shear.variants.Translocation;
import edu.umn.cs.kumarbio.shear.variants.Variant;

/**
 * SvHeterogeneityEstimator is used to estimate the heterogeneity for sets of
 * structural variants. It contains all of the logic for estimating
 * heterogeneity for different SV types.
 */
public class SvHeterogeneityEstimator {

  // The file path to the BAM alignment file to predict SV heterogeneity for
  private String  bamFilePath;

  // The file path to the FASTA reference file to predict SV heterogeneity for
  private String  faRefFilePath;

  // The instance of ShearSV that this heterogeneity estimator is called from
  private ShearSV shearSvInstance;

  /**
   * Creates a new SvHeterogeneityEstimator object.
   *
   * @param bamFilePath
   *          The file path to the BAM alignment file to predict SV
   *          heterogeneity for
   * @param faRefFilePath
   *          The file path to the FASTA reference file to predict SV
   *          heterogeneity for
   * @param shearSvInstance
   *          The instance of ShearSV that this heterogeneity estimator is
   *          called from
   */
  public SvHeterogeneityEstimator(String bamFilePath, String faRefFilePath, ShearSV shearSvInstance) {
    this.bamFilePath = bamFilePath;
    this.faRefFilePath = faRefFilePath;
    this.shearSvInstance = shearSvInstance;
  }

  /**
   * This estimates the heterogeneity for a set of structural variants.
   *
   * @param svs
   *          The set of structural variants to estimate the heterogeneity for
   */
  public void estimateHeterogeneity(TreeSet<Variant> svs) {

    // Loop through SVs
    for (Variant sv : svs) {

      // Estimated reference and variant depths for this SV
      double estVarDepth = 0d;
      double estRefDepth = 0d;

      // Breakpoints for the variant
      List<Location> bps = sv.getBreakpoints();

      // Breakpoint read information for this variant
      List<BreakpointReads> bpReads = new ArrayList<BreakpointReads>();

      // Estimate reference/variant depths for a deletion
      if (sv instanceof Deletion) {
        if (bps.size() != 2) {
          throw new ShearRuntimeException("Unexpected number of breakpoints for variant: " + sv + " (" + bps.size()
              + " breakpoints)");
        }
        BreakpointReads bp1reads = getBreakpointReads(bps.get(0), false);
        BreakpointReads bp2reads = getBreakpointReads(bps.get(1), false);
        bpReads.add(bp1reads);
        bpReads.add(bp2reads);
        estVarDepth = (double) bp1reads.getThreePrimeClippedReads() + (double) bp2reads.getFivePrimeClippedReads();
        estRefDepth = ((double) bp1reads.getSpanningReads() + (double) bp2reads.getSpanningReads()) / 2d;

        // Estimate reference/variant depths for an insertion
      } else if (sv instanceof Insertion) {
        if (bps.size() != 1) {
          throw new ShearRuntimeException("Unexpected number of breakpoints for variant: " + sv + " " + bps.size()
              + " breakpoints)");
        }
        BreakpointReads bp1reads = getBreakpointReads(bps.get(0), false);
        bpReads.add(bp1reads);
        estVarDepth =
            ((double) bp1reads.getThreePrimeClippedReads() + (double) bp1reads.getFivePrimeClippedReads()) / 2d;
        estRefDepth = bp1reads.getSpanningReads();

        // Estimate reference/variant depths for a tandem duplication
      } else if (sv instanceof TandemDuplication) {
        if (bps.size() != 2) {
          throw new ShearRuntimeException("Unexpected number of breakpoints for variant: " + sv + " " + bps.size()
              + " breakpoints)");
        }
        BreakpointReads bp1reads = getBreakpointReads(bps.get(0), false);
        BreakpointReads bp2reads = getBreakpointReads(bps.get(1), false);
        bpReads.add(bp1reads);
        bpReads.add(bp2reads);
        estVarDepth = (double) bp1reads.getFivePrimeClippedReads() + (double) bp2reads.getThreePrimeClippedReads();
        estRefDepth =
            Math.max(0d, ((double) bp1reads.getSpanningReads() + (double) bp2reads.getSpanningReads()) / 2d
                - estVarDepth);

        // Estimate reference/variant depths for an inversion
      } else if (sv instanceof Inversion) {
        if (bps.size() != 2) {
          throw new ShearRuntimeException("Unexpected number of breakpoints for variant: " + sv + " " + bps.size()
              + " breakpoints)");
        }
        BreakpointReads bp1reads = getBreakpointReads(bps.get(0), true);
        BreakpointReads bp2reads = getBreakpointReads(bps.get(1), true);
        bpReads.add(bp1reads);
        bpReads.add(bp2reads);
        estVarDepth =
            ((double) bp1reads.getFivePrimeClippedReads() + (double) bp1reads.getThreePrimeClippedReads()
                + bp2reads.getFivePrimeClippedReads() + bp2reads.getThreePrimeClippedReads()) / 2d;
        estRefDepth = ((double) bp1reads.getSpanningReads() + (double) bp2reads.getSpanningReads()) / 2d;

        // Estimate reference/variant depths for a translocation
      } else if (sv instanceof Translocation) {
        if (bps.size() != 2) {
          throw new ShearRuntimeException("Unexpected number of breakpoints for variant: " + sv + " " + bps.size()
              + " breakpoints)");
        }
        BreakpointReads bp1reads = getBreakpointReads(bps.get(0), false);
        BreakpointReads bp2reads = getBreakpointReads(bps.get(1), false);
        bpReads.add(bp1reads);
        bpReads.add(bp2reads);
        Translocation xloc = (Translocation) sv;
        switch (xloc.getTransType()) {
          case for3toFor5:
            estVarDepth = (double) bp1reads.getThreePrimeClippedReads() + (double) bp2reads.getFivePrimeClippedReads();
            break;
          case for3toRev5:
            estVarDepth = (double) bp1reads.getThreePrimeClippedReads() + (double) bp2reads.getThreePrimeClippedReads();
            break;
          case for5toRev3:
            estVarDepth = (double) bp1reads.getFivePrimeClippedReads() + (double) bp2reads.getFivePrimeClippedReads();
            break;
          default:
            throw new ShearRuntimeException("Illegal translocation type");
        }
        estRefDepth = ((double) bp1reads.getSpanningReads() + (double) bp2reads.getSpanningReads()) / 2d;

        // Throw error for unexpected Variant subclass
      } else {
        throw new ShearRuntimeException("Unexpected SV type encountered: " + sv);
      }

      // Set heterogeneity and breakpoint read information for SV and display
      double heterogeneity = estVarDepth / (estVarDepth + estRefDepth);
      sv.setHeterogeneity(Math.min(1d, heterogeneity));
      sv.setBpReads(bpReads);
      StringBuilder svAndHetString = new StringBuilder();
      svAndHetString.append("  " + sv + " ");
      for (int i = 11 + sv.toString().length(); i < 93; i++) {
        svAndHetString.append(".");
      }
      svAndHetString.append(String.format("%6.2f%%", 100d * sv.getHeterogeneity()));
      Shear.log(svAndHetString.toString(), shearSvInstance.isDebugMode());

      // Get the number of still-running ProgressMeterDaemon threads
      Set<Thread> threads = Thread.getAllStackTraces().keySet();
      int numProgressMeterDaemonThreads = 0;
      for (Thread t : threads) {
        if (t instanceof ProgressMeterDaemon) {
          numProgressMeterDaemonThreads++;
        }
      }

      // Wait for ProgressMeterDaemon threads to finish if there are too many of them
      if (numProgressMeterDaemonThreads >= 10) {
        try {
          Thread.sleep(ProgressMeterDaemon.DEFAULT_POLL_FREQUENCY_MILLISECONDS);
        } catch (InterruptedException e) {
          throw new ShearRuntimeException(
              "Thread interrupted while waiting for GATK progress meter daemons to close.",
              e);
        }
      }

    }

  }

  /**
   * Runs the BreakpointReadsWalker GATK command for the given breakpoint and
   * returns the resulting BreakpointReads object containing the clipped and
   * spanning reads at that breakpoint
   *
   * @param bp
   *          The breakpoint location to run BreakpointReadsWalker on (left
   *          side)
   * @param unclippedSpanningOnly
   *          Whether or not to require spanning reads to be completely
   *          unclipped at all locations
   * @return The BreakpointReads object containing the clipped and spanning
   *         reads at the specified breakpoint
   */
  private BreakpointReads getBreakpointReads(Location bp, boolean unclippedSpanningOnly) {

    // Temporary file to use
    String gatkOutputFilePath = shearSvInstance.getTmpFilePrefix() + "-" + bp + ".txt";

    // Prepare BreakpointReadsWalker GATK arguments
    String gatkArgs =
        String.format(
            "-T BreakpointReadsWalker -I %s -R %s -clipLocus %d -L %s:%d-%d -o %s",
            bamFilePath,
            faRefFilePath,
            bp.getCoord(),
            bp.getChr().getName(),
            bp.getCoord(),
            bp.getCoord() + 1,
            gatkOutputFilePath);
    if (unclippedSpanningOnly) {
      gatkArgs = gatkArgs + " -unclippedSpanningOnly";
    }
    if (shearSvInstance.isDebugMode()) {
      gatkArgs = gatkArgs + " -l INFO";
    } else {
      gatkArgs = gatkArgs + " -l ERROR";
    }

    // Run BreakpointReadsWalker GATK command
    RunGatkCommand.run(gatkArgs.split(" "), shearSvInstance.isDebugMode());

    // Number of clipped and spanning reads
    int fivePrimeClippedReads = 0;
    int threePrimeClippedReads = 0;
    int spanningReads = 0;

    // Parse BreakpointReadsWalker output and delete output file
    Scanner sc = null;
    File gatkOutputFile = new File(gatkOutputFilePath);
    try {
      sc = new Scanner(gatkOutputFile);
      String outputFileContent = sc.nextLine();
      if (sc.hasNextLine() || !outputFileContent.matches("^[0-9]+\t[0-9]+\t[0-9]+$")) {
        throw new ShearRuntimeException("Unexpected format in BreakpointReadsWalker output file: " + gatkOutputFilePath);
      }
      String[] tokens = outputFileContent.split("\t");
      fivePrimeClippedReads = Integer.valueOf(tokens[0]);
      threePrimeClippedReads = Integer.valueOf(tokens[1]);
      spanningReads = Integer.valueOf(tokens[2]);
    } catch (FileNotFoundException e) {
      throw new ShearRuntimeException("Problem reading from the BreakpointReadsWalker output file: "
          + gatkOutputFilePath, e);
    } finally {
      sc.close();
      gatkOutputFile.delete();
    }

    // Return BreakpointReads object
    return new BreakpointReads(fivePrimeClippedReads, threePrimeClippedReads, spanningReads);

  }

}
