/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio.shear.system;

import java.io.ByteArrayOutputStream;
import java.io.OutputStream;
import java.util.List;
import java.util.concurrent.FutureTask;

import edu.umn.cs.kumarbio.shear.Shear;
import edu.umn.cs.kumarbio.shear.ShearRuntimeException;

/**
 * RunSystemCommand is used to execute a system command and direct the output.
 * It contains only the static run methods, which take a ProcessBuilder object
 * and run it, possibly redirecting the output and error streams to a specified
 * OutputStream object, or returning output in a string. Any stream can also
 * have each line prepended with a tag to indicate the external program being
 * run.
 */
public class RunSystemCommand {

  /**
   * Runs the specified process and returns a string containing the output and
   * error stream results.
   *
   * @param pb
   *          The ProcessBuilder object containing the command to run
   * @return The output and error stream results of the process
   */
  public static String run(ProcessBuilder pb) {

    try {

      // Redirect error output to standard output
      pb.redirectErrorStream(true);

      // Set up OutputStream to collect result of process
      ByteArrayOutputStream result = new ByteArrayOutputStream();

      // Start the process
      Process p = pb.start();

      // Redirect the output stream in separate thread
      FutureTask<Boolean> redirector = new FutureTask<Boolean>(new StreamRedirector(p.getInputStream(), result));
      new Thread(redirector).start();

      // Wait for the output stream and for the process to finish
      redirector.get();
      p.waitFor();

      // Convert output to string and return
      return result.toString();

      // Catch errors
    } catch (Exception e) {
      throw new ShearRuntimeException("Error with running system command", e);
    }

  }

  /**
   * Runs the specified process and redirects the output and error streams,
   * possibly with tags to identify the external program. There are four
   * possibilities, depending on whether or not the debug tag is set and whether
   * or not the specified OutputStream object points to System.out. If the
   * target output is System.out and debug is OFF, then stdout and stderr are
   * ignored. If the target output is System.out and debug is NO, then stdout
   * and stderr are merged and a tag is used. If the target output is NOT
   * System.out and debug is OFF, then stdout is redirected to the specified
   * stream without a tag, and stderr is ignored. If the target output is NOT
   * System.out and debug is ON, then stdout is redirected to the specified
   * stream without a tag, and stderr is redirected to System.out with a tag.
   *
   * @param pb
   *          The ProcessBuilder object containing the command to run
   * @param outOutputStream
   *          The OutputStream to redirect the output stream to
   * @param tag
   *          The tag string to prepend to the redirected output stream content
   * @param debug
   *          Whether or not debugging mode is on
   */
  public static void run(ProcessBuilder pb, OutputStream outOutputStream, String tag, boolean debug) {

    // Print out the command being executed
    if (debug) {
      List<String> tokens = pb.command();
      StringBuilder command = new StringBuilder("  [DEBUG] SYSTEM-COMMAND:");
      for (int i = 0; i < tokens.size(); i++) {
        command.append(" " + tokens.get(i));
      }
      Shear.log(command.toString(), false);
    }

    // Tags to use for redirecting process's stdout and stderr
    String outTag = null;
    String errTag = null;

    // OutputStreams to use for redirecting process's stderr
    OutputStream errOutputStream = null;

    // If target output is standard out and we are in debug mode, redirect stderr to stdout and set an output tag
    if (outOutputStream == System.out && debug) {
      outTag = tag;
      pb.redirectErrorStream(true);

      // If target output is standard out and we are not in debug mode, set stdout OutputStream to null
    } else if (outOutputStream == System.out && !debug) {
      outOutputStream = null;

      // If target output is not standard out and we are in debug mode, set a tag for stderr and redirect it to System.out
    } else if (outOutputStream != System.out && debug) {
      errTag = tag;
      errOutputStream = System.out;
    }

    try {

      // Start the process
      Process p = pb.start();

      // Redirect the output stream
      FutureTask<Boolean> redirector =
          new FutureTask<Boolean>(new StreamRedirector(p.getInputStream(), outOutputStream, outTag));
      new Thread(redirector).start();

      // If not already redirected to stdout, redirect the error stream to  and wait for it
      if (!pb.redirectErrorStream()) {
        FutureTask<Boolean> errRedirector =
            new FutureTask<Boolean>(new StreamRedirector(p.getErrorStream(), errOutputStream, errTag));
        new Thread(errRedirector).start();
        errRedirector.get();
      }

      // Wait for the output stream and for the process to finish
      redirector.get();
      p.waitFor();

      // Print ending blank line
      if (debug) {
        System.out.println("[SHEAR]");
      }

      // Catch errors
    } catch (Exception e) {
      throw new ShearRuntimeException("Error with running system command", e);
    }

  }

}
