/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio.shear.system;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.concurrent.Callable;

import edu.umn.cs.kumarbio.shear.ShearRuntimeException;

/**
 * StreamRedirector is used to read the contents of an InputStream object and
 * write to an OutputStream object. This is used by the RunGatkCommand,
 * RunPicardCommand, and RunSystemCommand classes to run external programs, and
 * it implements Callable in order for it to be run concurrently (i.e. standard
 * error and standard output of an external program being redirected at the same
 * time). A tag can optionally be provided that will be prepended to each line
 * of the redirected content, which is useful for distinguishing the external
 * program whose output is being redirected. StreamRedirector will close the
 * InputStream object upon completion, but not the OutputStream object.
 */
public class StreamRedirector implements Callable<Boolean> {

  // The InputStream object to be redirected
  public InputStream  in;

  // The OutputStream object to write to
  public OutputStream out;

  // An optional tag string to prepend to the redirected content
  public String       tag;

  /**
   * Constructs a new StreamRedirector using an InputStream object and an
   * OutputStream object. If a non-null tag string is provided then content will
   * be redirected with prepended tags.
   *
   * @param in
   *          The InputStream object to be redirected
   * @param out
   *          The OutputStream object to write to
   * @param tag
   *          The tag string to prepend to the redirected content
   */
  public StreamRedirector(InputStream in, OutputStream out, String tag) {

    //Check that InputStream object is not null
    if (in == null) {
      throw new ShearRuntimeException("Null InputStream object provided to StreamRedirector");
    }

    // Set instance variables
    this.in = in;
    this.out = out;
    this.tag = tag;

  }

  /**
   * Constructs a new StreamRedirector using an InputStream object and an
   * OutputStream object. Content is purely redirected with no prepended tags.
   *
   * @param in
   *          The InputStream object to be redirected
   * @param out
   *          The OutputStream object to write to
   */
  public StreamRedirector(InputStream in, OutputStream out) {
    this(in, out, null);
  }

  /**
   * Redirects the stream. If a non-null tag string is provided in this
   * invocation of StreamRedirector then content will be redirected with a
   * prepended string of "[SHEAR]   [DEBUG] [TAG] " on every line, where TAG is
   * the specified tag string. The InputStream object will be closed upon
   * completion, but not the OutputStream object. OutputStream can be null to
   * ignore output.
   *
   * @throws IOException
   *           If there are any problems reading or writing to the streams
   */
  @Override
  public Boolean call() throws IOException {

    // Buffer to store input from the InputStream object
    byte[] buf = new byte[1024];

    // Keep track of number of bytes read from the InputStream object
    int len;

    // If OutputStream object is null, ignore all content
    if (out == null) {
      while ((len = in.read(buf)) > 0) {
      }

      // If tag is not null, redirect with prepended tag
    } else if (tag != null) {

      // Whether or not we are at the start of a new line (initialize to true)
      boolean newLine = true;

      // Loop through content in InputStream object
      while ((len = in.read(buf)) > 0) {

        // Process current buffer of bytes from InputStream object
        for (int i = 0; i < len; i++) {

          // Write the tags if newLine is true
          if (newLine) {
            out.write(new String("[SHEAR]   [DEBUG] [" + tag + "] ").getBytes());
            newLine = false;
          }

          // Write the current byte
          out.write(buf[i]);

          // If we just wrote a "\n" set newLine = true for next time
          if (buf[i] == new String("\n").getBytes()[0]) {
            newLine = true;
          }

        }

      }

      // If tag is null, redirect normally
    } else {
      while ((len = in.read(buf)) > 0) {
        out.write(buf, 0, len);
      }
    }

    // Closes the InputStream object
    in.close();

    return true;

  }

}
