/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio.shear.variantcallers;

import java.io.File;
import java.io.FileNotFoundException;
import java.util.ArrayList;
import java.util.List;
import java.util.Scanner;
import java.util.TreeSet;

import edu.umn.cs.kumarbio.Genome;
import edu.umn.cs.kumarbio.shear.Shear;
import edu.umn.cs.kumarbio.shear.ShearRuntimeException;
import edu.umn.cs.kumarbio.shear.system.RunSystemCommand;
import edu.umn.cs.kumarbio.shear.variants.Deletion;
import edu.umn.cs.kumarbio.shear.variants.IllegalVariantException;
import edu.umn.cs.kumarbio.shear.variants.Inversion;
import edu.umn.cs.kumarbio.shear.variants.Location;
import edu.umn.cs.kumarbio.shear.variants.TandemDuplication;
import edu.umn.cs.kumarbio.shear.variants.Translocation;
import edu.umn.cs.kumarbio.shear.variants.Variant;

/**
 * This is a wrapper for the CREST SV caller. This class extends the abstract
 * class VariantCaller.
 */
public class CrestWrapper extends VariantCaller {

  // The Genome object to use for this instance of CREST
  private Genome   gen;

  // Whether or not debugging mode is active
  private boolean  debug;

  // The average read length for the data to process
  private int      readLength;

  // The directory that contains the CREST scripts to use
  private String   crestDir;

  // The prefix to use for log and output files for the GFServer
  private String   gfServerPrefix;

  // The *.2bit file used by the GFServer
  private String   gfServerTwoBitFile;

  // The GFServer object to use for this instance of CREST
  private GFServer gfServer;

  /**
   * Creates a new CrestWrapper using the specified genome. The FASTA reference
   * file associated the genome is required to have an accompanying *.fai index
   * file. Also requires the directory that contains the CREST scripts, as well
   * as the average read length of the data to process and information required
   * to start up the GFServer.
   *
   * @param gen
   *          The Genome object to use for this instance of CREST
   * @param debug
   *          Whether or not debugging mode is active
   * @param readLength
   *          The average read length for the data to process
   * @param crestDir
   *          The directory that contains the CREST scripts to use
   * @param gfServerPrefix
   *          The prefix to use for log and output files for the GFServer
   * @param gfServerTwoBitFile
   *          The *.2bit file used by the GFServer
   */
  public CrestWrapper(
      Genome gen, boolean debug, int readLength, String crestDir, String gfServerPrefix, String gfServerTwoBitFile) {

    // Set instance variables
    this.gen = gen;
    this.debug = debug;
    this.readLength = readLength;
    this.crestDir = crestDir;
    this.gfServerPrefix = gfServerPrefix;
    this.gfServerTwoBitFile = gfServerTwoBitFile;

    // Initialize GFServer
    gfServer = initializeGFServer();

  }

  /**
   * Sets up an instance of GFServer to be used for this instance of CREST.
   *
   * @return The GFServer object to use for this instance of CREST after it is
   *         set up
   */
  private GFServer initializeGFServer() {

    try {

      // Initialize GFServer object
      GFServer gfServer = new GFServer(gfServerPrefix, gfServerTwoBitFile);

      // Start GFServer
      if (debug) {
        Shear.log("[DEBUG] Using port " + gfServer.getPortNum() + " for gfServer", false);
        Shear.log("[DEBUG] Starting gfServer...", false);
      }
      gfServer.start();

      // Wait for GFServer to be ready
      if (debug) {
        Shear.log("[DEBUG] Waiting for gfServer to be ready...", false);
      }
      boolean gfServerStarted = gfServer.waitForReady();

      // Loop for every GFServer that was not successfully started
      while (!gfServerStarted) {

        // Throw error if *.2bit file is invalid
        if (!gfServer.isValidTwoBitFile()) {
          throw new ShearRuntimeException("Not a valid two-bit reference file: " + gfServerTwoBitFile);
        }

        // If bad port number was used, clear log files and close GFServer
        if (debug) {
          Shear.log("[DEBUG] Port already in use. Restarting gfServer...", false);
        }
        gfServer.clearLogFiles();
        gfServer.close();

        // Initialize new GFServer object starting at the next port number after the failed one
        gfServer = new GFServer(gfServerPrefix, gfServerTwoBitFile, gfServer.getPortNum() + 1);

        // Start GFServer
        if (debug) {
          Shear.log("[DEBUG] Using port " + gfServer.getPortNum() + " for gfServer", false);
          Shear.log("[DEBUG] Starting gfServer...", false);
        }
        gfServer.start();

        // Wait for new GFServer to be ready
        if (debug) {
          Shear.log("[DEBUG] Waiting for gfServer to be ready...", false);
        }
        gfServerStarted = gfServer.waitForReady();

      }

      // Print debugging log message that GFServer is ready
      if (debug) {
        Shear.log("[DEBUG] gfServer is ready!", false);
      }

      // Return GFServer object
      return gfServer;

      // Catch possible errors
    } catch (FileNotFoundException e) {
      throw new ShearRuntimeException("Unable to create GFServer to be used by CREST", e);
    } catch (InterruptedException e) {
      throw new ShearRuntimeException("Unable to create GFServer to be used by CREST", e);
    }

  }

  @Override
  public void close() {

    // Display log message
    if (debug) {
      Shear.log("[DEBUG] Closing gfServer...", false);
    }

    // Close GFServer
    try {
      gfServer.close();
    } catch (Exception e) {
      throw new ShearRuntimeException("Error while trying to close down GFServer");
    }

    // Delete log files
    if (!debug) {
      gfServer.clearLogFiles();
    }

  }

  @Override
  public String runVariantCaller(String bamFile, String outputPrefix, List<String> options) {

    // Output files to use
    String crestSclipFilePath = outputPrefix + ".sclip.txt";
    String crestCoverFilePath = outputPrefix + ".cover";
    String crestPredFilePath = outputPrefix + ".predSV.txt";

    // Delete files that may be present from a previous run
    new File(crestSclipFilePath).delete();
    new File(crestCoverFilePath).delete();
    new File(crestPredFilePath).delete();

    // Run CREST's extractSClip.pl script
    ProcessBuilder extractSClipProcess =
        new ProcessBuilder(
            "perl",
            "-I" + crestDir,
            crestDir + "/extractSClip.pl",
            "-p",
            outputPrefix,
            "-i",
            bamFile,
            "--ref_genome",
            gen.getFaFilePath());
    RunSystemCommand.run(extractSClipProcess, System.out, "CREST", debug);

    // Prepare list of command and arguments for CREST.pl script
    ArrayList<String> crestCommandAndArgs = new ArrayList<String>();
    crestCommandAndArgs.add("perl");
    crestCommandAndArgs.add("-I" + crestDir);
    crestCommandAndArgs.add(crestDir + "/CREST.pl");
    crestCommandAndArgs.add("--read_len");
    crestCommandAndArgs.add(String.valueOf(readLength));
    crestCommandAndArgs.add("--max_rep_cover");
    crestCommandAndArgs.add("1000000");
    for (int i = 0; i < options.size(); i++) {
      crestCommandAndArgs.add(options.get(i));
    }
    crestCommandAndArgs.add("-p");
    crestCommandAndArgs.add(outputPrefix);
    crestCommandAndArgs.add("-f");
    crestCommandAndArgs.add(crestCoverFilePath);
    crestCommandAndArgs.add("-d");
    crestCommandAndArgs.add(bamFile);
    crestCommandAndArgs.add("--ref_genome");
    crestCommandAndArgs.add(gen.getFaFilePath());
    crestCommandAndArgs.add("-t");
    crestCommandAndArgs.add(gfServer.getTwoBitFile());
    crestCommandAndArgs.add("--blatserver");
    crestCommandAndArgs.add("localhost");
    crestCommandAndArgs.add("--blatport");
    crestCommandAndArgs.add(String.valueOf(gfServer.getPortNum()));

    // Run CREST's CREST.pl script
    ProcessBuilder crestProcess = new ProcessBuilder(crestCommandAndArgs);
    RunSystemCommand.run(crestProcess, System.out, "CREST", debug);

    // Clean up extractSClip.pl output files if not in debug mode
    if (!debug) {
      new File(crestSclipFilePath).delete();
      new File(crestCoverFilePath).delete();
    }

    // Return output file path
    return crestPredFilePath;

  }

  @Override
  public TreeSet<Variant> getVariants(String variantCallerOutputFile) {

    // Note that this getVariants implementation DOES NOT calculate heterogeneity percentages for variants

    // Initialize set of variants to return
    TreeSet<Variant> variants = new TreeSet<Variant>();

    // Open up a scanner for predSV.txt output file and read from it
    Scanner sc = null;
    try {
      sc = new Scanner(new File(variantCallerOutputFile));
      while (sc.hasNextLine()) {
        String line = sc.nextLine();
        String[] tokens = line.split("\t");

        // Validate that line has expected format
        if (tokens.length >= 9 && tokens[1].matches("^[0-9]+$") && tokens[5].matches("^[0-9]+$")
            && tokens[2].matches("^\\+|-$") && tokens[6].matches("^\\+|-$")) {

          // Get relevant predSV.txt line information
          String chr1 = tokens[0];
          long loc1 = Long.valueOf(tokens[1]);
          String strand1 = tokens[2];
          String chr2 = tokens[4];
          long loc2 = Long.valueOf(tokens[5]);
          String strand2 = tokens[6];
          String svType = tokens[8];

          // The variant object
          Variant v = null;

          try {

            // Create Deletion
            if (svType.equals("DEL") && chr1.equals(chr2) && loc1 < loc2 - 1 && strand1.equals("+")
                && strand2.equals("+")) {
              v = new Deletion(new Location(gen.getChromosome(chr1), loc1 + 1), loc2 - loc1 - 1, false);

              // Create Tandem Duplication
            } else if (svType.equals("INS") && chr1.equals(chr2) && loc1 >= loc2 && strand1.equals("+")
                && strand2.equals("+")) {
              v = new TandemDuplication(new Location(gen.getChromosome(chr2), loc2), loc1 - loc2 + 1);

              // Create Inversion (type 1, + > -)
            } else if (svType.equals("INV") && chr1.equals(chr2) && loc1 < loc2 - 1 && strand1.equals("+")
                && strand2.equals("-")) {
              v = new Inversion(new Location(gen.getChromosome(chr1), loc1 + 1), loc2 - loc1 - 1);

              // Create Inversion (type 2, - > +)
            } else if (svType.equals("INV") && chr1.equals(chr2) && loc1 >= loc2 && strand1.equals("-")
                && strand2.equals("+")) {
              v = new Inversion(new Location(gen.getChromosome(chr2), loc2), loc1 - loc2 + 1);

              // Create Intra-Chromosome Translocation (type 1, + > -)
            } else if (svType.equals("ITX") && chr1.equals(chr2) && strand1.equals("+") && strand2.equals("-")) {
              v =
                  new Translocation(new Location(gen.getChromosome(chr1), loc1), new Location(
                      gen.getChromosome(chr2),
                      loc2), Translocation.Type.for3toRev5);

              // Create Intra-Chromosome Translocation (type 2, - > +)
            } else if (svType.equals("ITX") && chr1.equals(chr2) && strand1.equals("-") && strand2.equals("+")) {
              v =
                  new Translocation(new Location(gen.getChromosome(chr2), loc2), new Location(
                      gen.getChromosome(chr1),
                      loc1), Translocation.Type.for5toRev3);

              // Create Inter-Chromosome Translocation (type 1, + > -)
            } else if (svType.equals("CTX") && !chr1.equals(chr2) && strand1.equals("+") && strand2.equals("-")) {
              v =
                  new Translocation(new Location(gen.getChromosome(chr1), loc1), new Location(
                      gen.getChromosome(chr2),
                      loc2), Translocation.Type.for3toRev5);

              // Create Inter-Chromosome Translocation (type 2, - > +)
            } else if (svType.equals("CTX") && !chr1.equals(chr2) && strand1.equals("-") && strand2.equals("+")) {
              v =
                  new Translocation(new Location(gen.getChromosome(chr2), loc2), new Location(
                      gen.getChromosome(chr1),
                      loc1), Translocation.Type.for5toRev3);

              // Create Inter-Chromosome Translocation (type 3, + > +)
            } else if (svType.equals("CTX") && !chr1.equals(chr2) && strand1.equals("+") && strand2.equals("+")) {
              v =
                  new Translocation(new Location(gen.getChromosome(chr1), loc1), new Location(
                      gen.getChromosome(chr2),
                      loc2), Translocation.Type.for3toFor5);

              // Skip over lines with unexpected format
            } else {
              Shear.log("WARNING: Skipping unexpected event in CREST output (" + tokens[0] + "\t" + tokens[1] + "\t"
                  + tokens[2] + "\t" + tokens[3] + "\t" + tokens[4] + "\t" + tokens[5] + "\t" + tokens[6] + "\t"
                  + tokens[7] + "\t" + tokens[8] + ")", debug);
              continue;

            }

            // Skip over line that caused IllegalVariantException error
          } catch (IllegalVariantException e) {
            Shear.log("[DEBUG] WARNING: Skipping illegal variant in CREST output (" + tokens[0] + "\t" + tokens[1]
                + "\t" + tokens[2] + "\t" + tokens[3] + "\t" + tokens[4] + "\t" + tokens[5] + "\t" + tokens[6] + "\t"
                + tokens[7] + "\t" + tokens[8] + ")", debug);
            continue;

          }

          // Add variant to return set
          variants.add(v);

          // Throw error for unexpected predSV.txt format
        } else {
          throw new ShearRuntimeException("Unexpected format in predSV.txt output file '" + variantCallerOutputFile
              + "' from CREST");
        }

      }

      // Catch error from setting up Scanner object for predSV.txt output
    } catch (FileNotFoundException e) {
      throw new ShearRuntimeException("Unable to locate predSV.txt output file '" + variantCallerOutputFile
          + "' from CREST", e);

    } finally {

      // Close Scanner object
      if (sc != null) {
        sc.close();
      }

      // Clean up predSV.txt output file if not in debug mode
      if (!debug) {
        new File(variantCallerOutputFile).delete();
      }

    }

    // Return variants
    return variants;

  }

}
