/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio.shear.variants;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import edu.umn.cs.kumarbio.shear.ShearRuntimeException;
import edu.umn.cs.kumarbio.shear.heterogeneity.BreakpointReads;

/**
 * Represents a deletion in a genome.
 */
public class Deletion extends Variant {

  // The size of this deletion
  private long size;

  /**
   * Creates a new Deletion object.
   *
   * @param loc
   *          The location of this deletion (i.e. first deleted base in the
   *          chromosome)
   * @param size
   *          The size of this deletion
   * @param isIndel
   *          Whether or not this deletion is an indel (i.e. small insertion or
   *          deletion)
   */
  public Deletion(Location loc, long size, boolean isIndel) {
    super(loc, isIndel);

    // Parse arguments and throw errors
    if (size < 1) {
      throw new IllegalVariantException("Invalid size specified for deletion (non-positive value): " + size + "bp");
    } else if (loc.getCoord() + size - 1 > loc.getChr().getLength()) {
      throw new IllegalVariantException("Invalid size specified for deletion (too large for chromosome): "
          + loc.getChr().getName() + ":" + loc + " (" + size + "bp)");
    }

    // Initialize variables
    this.size = size;

  }

  /**
   * Returns the size of this deletion.
   *
   * @return The size of this deletion
   */
  public long getSize() {
    return size;
  }

  /**
   * Returns the end location of this deletion (i.e. last deleted base in the
   * chromosome).
   *
   * @return The end location of this deletion (i.e. last deleted base in the
   *         chromosome)
   */
  public Location getEndLoc() {
    return new Location(getChr(), getEndCoord());
  }

  /**
   * Returns the end coordinate of this deletion (i.e. last deleted base in the
   * chromosome).
   *
   * @return The end coordinate of this deletion (i.e. last deleted base in the
   *         chromosome)
   */
  public long getEndCoord() {
    return getCoord() + getSize() - 1;
  }

  /**
   * Returns the string of deleted bases for this deletion.
   *
   * @return The string of deleted bases for this deletion
   */
  public String getDeletedBases() {
    try {
      return getChr().getSequence(getCoord(), getEndCoord());
    } catch (IOException e) {
      throw new ShearRuntimeException("Error reading from the FASTA reference file", e);
    }
  }

  @Override
  public List<Location> getBreakpoints() {

    // Breakpoints are just before the deletion and last base of deletion
    ArrayList<Location> bps = new ArrayList<Location>();
    bps.add(new Location(getChr(), getCoord() - 1));
    bps.add(getEndLoc());
    return bps;

  }

  @Override
  public String getSdiString() {
    return formatSdiLine(getLoc(), getDeletedBases(), "-");
  }

  @Override
  public String getVariantIdString() {
    String variantIdString = "DEL\t" + getLoc() + "\t" + getSize() + "\t " + getHeterogeneity();
    if (getBpReads() != null) {
      variantIdString += "\t [";
      for (BreakpointReads bpReads : getBpReads()) {
        variantIdString += bpReads + " ";
      }
      variantIdString += "]";
    }
    variantIdString += "\t" + isIndel();
    return variantIdString;
  }

  @Override
  public boolean equals(Object obj) {
    if (obj == null || !(obj instanceof Deletion)) {
      return false;
    } else {
      Deletion other = (Deletion) obj;
      return getLoc().equals(other.getLoc()) && getHeterogeneity() == other.getHeterogeneity()
          && isIndel() == other.isIndel() && getSize() == other.getSize();
    }
  }

  @Override
  public String toString() {
    return String.format("%-6s%s:%d-%d (%d bp)", "DEL", getChr().getName(), getCoord(), getEndCoord(), getSize());
  }

}
