/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio.shear.variants;

import java.util.ArrayList;
import java.util.List;

import edu.umn.cs.kumarbio.shear.heterogeneity.BreakpointReads;

/**
 * Represents an insertion in a genome.
 */
public class Insertion extends Variant {

  // The string of inserted bases for this insertion
  private String insertedBases;

  /**
   * Creates a new Insertion object.
   *
   * @param loc
   *          The location of this insertion (i.e. position in the chromosome
   *          that will now appear directly after the inserted bases)
   * @param insertedBases
   *          The string of inserted bases for this insertion
   * @param isIndel
   *          Whether or not this insertion is an indel (i.e. small insertion or
   *          deletion)
   */
  public Insertion(Location loc, String insertedBases, boolean isIndel) {
    super(loc, isIndel);

    // Parse arguments and throw errors
    if (insertedBases == null) {
      throw new IllegalVariantException("Null inserted bases string specified for insertion");
    } else if (insertedBases.length() == 0) {
      throw new IllegalVariantException("Empty inserted bases string specified for insertion");
    } else if (!insertedBases.matches("^[ACGTNacgtn]+$")) {
      if (insertedBases.length() > 10) {
        throw new IllegalVariantException("Illegal inserted bases string specified for insertion: "
            + insertedBases.substring(0, 10) + "...");
      } else {
        throw new IllegalVariantException("Illegal inserted bases string specified for insertion: " + insertedBases);
      }
    }

    // Initialize variables
    this.insertedBases = insertedBases;

  }

  /**
   * Returns the string of inserted bases for this insertion.
   *
   * @return The string of inserted bases for this insertion
   */
  public String getInsertedBases() {
    return insertedBases;
  }

  /**
   * Returns the size of this insertion.
   *
   * @return The size of this insertion
   */
  public long getSize() {
    return getInsertedBases().length();
  }

  @Override
  public List<Location> getBreakpoints() {

    // Breakpoint is just before the insertion
    ArrayList<Location> bps = new ArrayList<Location>();
    bps.add(new Location(getChr(), getCoord() - 1));
    return bps;

  }

  @Override
  public String getSdiString() {
    return formatSdiLine(getLoc(), "-", getInsertedBases());
  }

  @Override
  public String getVariantIdString() {
    String variantIdString = "INS\t" + getLoc() + "\t" + getInsertedBases() + "\t " + getHeterogeneity();
    if (getBpReads() != null) {
      variantIdString += "\t [";
      for (BreakpointReads bpReads : getBpReads()) {
        variantIdString += bpReads + " ";
      }
      variantIdString += "]";
    }
    variantIdString += "\t" + isIndel();
    return variantIdString;
  }

  @Override
  public boolean equals(Object obj) {
    if (obj == null || !(obj instanceof Insertion)) {
      return false;
    } else {
      Insertion other = (Insertion) obj;
      return getLoc().equals(other.getLoc()) && getHeterogeneity() == other.getHeterogeneity()
          && isIndel() == other.isIndel() && getInsertedBases().equals(other.getInsertedBases());
    }
  }

  @Override
  public String toString() {
    return String.format("%-6s%s:%d (%d bp)", "INS", getChr().getName(), getCoord(), getInsertedBases().length());
  }

}
