/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio.shear.variants;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import edu.umn.cs.kumarbio.Genome;
import edu.umn.cs.kumarbio.shear.ShearRuntimeException;
import edu.umn.cs.kumarbio.shear.heterogeneity.BreakpointReads;

/**
 * Represents an inversion in a genome.
 */
public class Inversion extends Variant {

  // The size of this inversion
  private long size;

  /**
   * Creates a new Inversion object.
   *
   * @param loc
   *          The location of this inversion (i.e. first affected base in the
   *          chromosome)
   * @param size
   *          The size of this inversion
   */
  public Inversion(Location loc, long size) {
    super(loc);

    // Parse arguments and throw errors
    if (size < 1) {
      throw new IllegalVariantException("Invalid size specified for inversion (non-positive value): " + size + "bp");
    } else if (loc.getCoord() + size - 1 > loc.getChr().getLength()) {
      throw new IllegalVariantException("Invalid size specified for inversion (too large for chromosome): "
          + loc.getChr().getName() + ":" + loc + " (" + size + "bp)");
    }

    // Initialize variables
    this.size = size;

  }

  /**
   * Returns the size of this inversion.
   *
   * @return The size of this inversion
   */
  public long getSize() {
    return size;
  }

  /**
   * Returns the end location of this inversion (i.e. last affected base in the
   * chromosome).
   *
   * @return The end location of this inversion (i.e. last affected base in the
   *         chromosome)
   */
  public Location getEndLoc() {
    return new Location(getChr(), getEndCoord());
  }

  /**
   * Returns the end coordinate of this inversion (i.e. last affected base in
   * the chromosome).
   *
   * @return The end coordinate of this inversion (i.e. last affected base in
   *         the chromosome)
   */
  public long getEndCoord() {
    return getCoord() + getSize() - 1;
  }

  /**
   * Returns the string of affected bases (before inverting) for this inversion.
   *
   * @return The string of affected bases (before inverting) for this inversion
   */
  public String getAffectedBases() {
    try {
      return getChr().getSequence(getCoord(), getEndCoord());
    } catch (IOException e) {
      throw new ShearRuntimeException("Error reading from the FASTA reference file", e);
    }
  }

  /**
   * Returns the string of inverted bases for this inversion.
   *
   * @return The string of inverted bases for this inversion
   */
  public String getInvertedBases() {
    return Genome.rc(getAffectedBases());
  }

  @Override
  public List<Location> getBreakpoints() {

    // Breakpoints are just before the inversion and last base of inversion
    ArrayList<Location> bps = new ArrayList<Location>();
    bps.add(new Location(getChr(), getCoord() - 1));
    bps.add(getEndLoc());
    return bps;

  }

  @Override
  public String getSdiString() {
    return formatSdiLine(getLoc(), getAffectedBases(), getInvertedBases());
  }

  @Override
  public String getVariantIdString() {
    String variantIdString = "INV\t" + getLoc() + "\t" + getSize() + "\t " + getHeterogeneity();
    if (getBpReads() != null) {
      variantIdString += "\t [";
      for (BreakpointReads bpReads : getBpReads()) {
        variantIdString += bpReads + " ";
      }
      variantIdString += "]";
    }
    variantIdString += "\t" + isIndel();
    return variantIdString;
  }

  @Override
  public boolean equals(Object obj) {
    if (obj == null || !(obj instanceof Inversion)) {
      return false;
    } else {
      Inversion other = (Inversion) obj;
      return getLoc().equals(other.getLoc()) && getHeterogeneity() == other.getHeterogeneity()
          && isIndel() == other.isIndel() && getSize() == other.getSize();
    }
  }

  @Override
  public String toString() {
    return String.format("%-6s%s:%d-%d (%d bp)", "INV", getChr().getName(), getCoord(), getEndCoord(), getSize());
  }

}
