/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio.shear.variants;

import edu.umn.cs.kumarbio.Chromosome;
import edu.umn.cs.kumarbio.shear.ShearRuntimeException;

/**
 * Represents a location in a chromosome.
 */
public class Location implements Comparable<Location> {

  // The chromosome of this location
  private Chromosome chr;

  // The coordinate of this location
  private long       coord;

  /**
   * Creates a new Location object based on the specified chromosome and
   * coordinate.
   *
   * @param chr
   *          The chromosome of this location
   * @param coord
   *          The coordinate of this location
   */
  public Location(Chromosome chr, long coord) {

    // Parse arguments and throw errors
    if (chr == null) {
      throw new IllegalVariantException("Null chromosome specified for location");
    } else if (coord < 1 || coord > chr.getLength()) {
      throw new IllegalVariantException("Invalid coordinate specified for location: " + chr.getName() + ":" + coord);
    }

    // Set instance variables
    this.chr = chr;
    this.coord = coord;

  }

  /**
   * Returns the chromosome of this location.
   *
   * @return The chromosome of this location
   */
  public Chromosome getChr() {
    return chr;
  }

  /**
   * Returns the coordinate of this location.
   *
   * @return The coordinate of this location
   */
  public long getCoord() {
    return coord;
  }

  @Override
  public int compareTo(Location other) {

    // Return 0 if variants are equal
    if (equals(other)) {
      return 0;
    }

    // Return whichever has first chromosome
    if (getChr().compareTo(other.getChr()) < 0) {
      return -1;
    } else if (getChr().compareTo(other.getChr()) > 0) {
      return +1;
    }

    // If chromosomes are the same, return whichever has first loc
    if (getCoord() < other.getCoord()) {
      return -1;
    } else if (getCoord() > other.getCoord()) {
      return +1;

      // Throw error for exact same location (should be unreachable code)
    } else {
      throw new ShearRuntimeException("Unexpected behavior while comparing order of Location objects"
          + " (Same chromosome and coordinate, but objects not equal)");
    }

  }

  @Override
  public boolean equals(Object obj) {
    if (obj == null || !(obj instanceof Location)) {
      return false;
    } else {
      Location other = (Location) obj;
      return getChr().equals(other.getChr()) && getCoord() == other.getCoord();
    }
  }

  @Override
  public String toString() {
    return getChr().getName() + ":" + getCoord();
  }

}
