/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio.shear.variants;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import edu.umn.cs.kumarbio.shear.ShearRuntimeException;
import edu.umn.cs.kumarbio.shear.heterogeneity.BreakpointReads;

/**
 * Represents a tandem duplication in a genome.
 */
public class TandemDuplication extends Variant {

  // The size of this tandem duplication (i.e. how many of the original bases are duplicated)
  private long size;

  /**
   * Creates a new TandemDuplication object.
   *
   * @param loc
   *          The location of this tandem duplication (i.e. first base on the
   *          chromosome that is duplicated)
   * @param size
   *          The size of this tandem duplication (i.e. how many of the original
   *          bases are duplicated)
   */
  public TandemDuplication(Location loc, long size) {
    super(loc);

    // Parse arguments and throw errors
    if (size < 1) {
      throw new IllegalVariantException("Invalid size specified for tandem duplication (non-positive value): " + size
          + "bp");
    } else if (loc.getCoord() + size - 1 > loc.getChr().getLength()) {
      throw new IllegalVariantException("Invalid size specified for tandem duplication (too large for chromosome): "
          + loc.getChr().getName() + ":" + loc + " (" + size + "bp)");
    }

    // Initialize variables
    this.size = size;

  }

  /**
   * Returns the size of this tandem duplication (i.e. how many of the original
   * bases are duplicated).
   *
   * @return The size of this tandem duplication (i.e. how many of the original
   *         bases are duplicated)
   */
  public long getSize() {
    return size;
  }

  /**
   * Returns the end location of this tandem duplication (i.e. last base on the
   * chromosome that is duplicated).
   *
   * @return The end location of this tandem duplication (i.e. last base on the
   *         chromosome that is duplicated)
   */
  public Location getEndLoc() {
    return new Location(getChr(), getEndCoord());
  }

  /**
   * Returns the end coordinate of this tandem duplication (i.e. last base on
   * the chromosome that is duplicated).
   *
   * @return The end coordinate of this tandem duplication (i.e. last base on
   *         the chromosome that is duplicated)
   */
  public long getEndCoord() {
    return getCoord() + getSize() - 1;
  }

  /**
   * Returns the string of duplicated bases for this tandeom duplication.
   *
   * @return The string of duplicated bases for this tandeom duplication
   */
  public String getDuplicatedBases() {
    try {
      return getChr().getSequence(getCoord(), getEndCoord());
    } catch (IOException e) {
      throw new ShearRuntimeException("Error reading from the FASTA reference file", e);
    }
  }

  @Override
  public List<Location> getBreakpoints() {

    // Breakpoints are just before the tandem duplication and last base of tandem duplication
    ArrayList<Location> bps = new ArrayList<Location>();
    bps.add(new Location(getChr(), getCoord() - 1));
    bps.add(getEndLoc());
    return bps;

  }

  @Override
  public String getSdiString() {
    return formatSdiLine(getLoc(), "-", getDuplicatedBases());
  }

  @Override
  public String getVariantIdString() {
    String variantIdString = "TDUP\t" + getLoc() + "\t" + getSize() + "\t " + getHeterogeneity();
    if (getBpReads() != null) {
      variantIdString += "\t [";
      for (BreakpointReads bpReads : getBpReads()) {
        variantIdString += bpReads + " ";
      }
      variantIdString += "]";
    }
    variantIdString += "\t" + isIndel();
    return variantIdString;
  }

  @Override
  public boolean equals(Object obj) {
    if (obj == null || !(obj instanceof TandemDuplication)) {
      return false;
    } else {
      TandemDuplication other = (TandemDuplication) obj;
      return getLoc().equals(other.getLoc()) && getHeterogeneity() == other.getHeterogeneity()
          && isIndel() == other.isIndel() && getSize() == other.getSize();
    }
  }

  @Override
  public String toString() {
    return String.format("%-6s%s:%d-%d (%d bp)", "TDUP", getChr().getName(), getCoord(), getEndCoord(), getSize());
  }

}
