/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio.shear.variants;

import java.util.ArrayList;
import java.util.List;

import edu.umn.cs.kumarbio.Chromosome;
import edu.umn.cs.kumarbio.shear.ShearRuntimeException;
import edu.umn.cs.kumarbio.shear.heterogeneity.BreakpointReads;

/**
 * Represents a translocation in a genome. This is represented by two locations
 * which are joined. The first location should refer to a base on the forward
 * strand. The second location specifies the base which is joined to the first
 * location, but this can refer to the base on the forward strand (3' to 5' end
 * join), to the 5' side of a base on the reverse strand joined through the 3'
 * end of the first location, or to the 3' side of a base on the reverse strand
 * joined through the 5' of the first location.
 */
public class Translocation extends Variant {

  // The second location that this translocation involves
  private Location loc2;

  // The type of translocation join
  private Type     transType;

  /**
   * This represents the possible types of translocation joins. The first
   * location in the Translocation class always refers to a base on the forward
   * strand. There are three possibilities for how this base is joined with the
   * second location:
   * <ul>
   * <li>for3toFor5: The 3' end of the forward base at the first loc is joined
   * to the 5' end of the forward base at the second loc</li>
   * <li>for3toRev5: The 3' end of the forward base at the first loc is joined
   * to the 5' end of the reverse base at the second loc</li>
   * <li>for5toRev3: The 5' end of the forward base at the first loc is joined
   * to the 3' end of the reverse base at the second loc</li>
   * </ul>
   */
  public enum Type {
    for3toFor5, for3toRev5, for5toRev3
  };

  /**
   * Create a new Translocation object.
   *
   * @param loc1
   *          The first location that this translocation involves
   * @param loc2
   *          The second location that this translocation involves
   * @param transType
   *          Whether or not this translocation joins the forward strand at the
   *          first location to the reverse strand at the second location
   */
  public Translocation(Location loc1, Location loc2, Type transType) {
    super(loc1);

    // Parse arguments and throw errors
    if (loc2 == null) {
      throw new IllegalVariantException("Null second location specified for translocation");
    } else if (loc1.getChr().getGenome() != loc2.getChr().getGenome()) {
      throw new IllegalVariantException(
          "Incompatable chromosomes specified for translocation (different Genome objects): " + loc1.getChr().getName()
              + " / " + loc2.getChr().getName());
    }

    // Initialize variables
    this.loc2 = loc2;
    this.transType = transType;

  }

  /**
   * Returns the second chromosome that this translocation involves.
   *
   * @return The second chromosome that this translocation involves
   */
  public Chromosome getSecondChr() {
    return getSecondLoc().getChr();
  }

  /**
   * Returns the second coordinate that this translocation involves.
   *
   * @return The second coordinate that this translocation involves.
   */
  public long getSecondCoord() {
    return getSecondLoc().getCoord();
  }

  /**
   * Returns the second location that this translocation involves.
   *
   * @return The second location that this translocation involves
   */
  public Location getSecondLoc() {
    return loc2;
  }

  /**
   * Returns the type of translocation join that this translocation involves
   *
   * @return The type of translocation join that this translocation involves
   */
  public Type getTransType() {
    return transType;
  }

  @Override
  public List<Location> getBreakpoints() {

    // Breakpoints depend on the type of translocation
    ArrayList<Location> bps = new ArrayList<Location>();
    if (getTransType() == Type.for3toFor5) {
      bps.add(getLoc());
      bps.add(new Location(getSecondChr(), getSecondCoord() - 1));
    } else if (getTransType() == Type.for3toRev5) {
      bps.add(getLoc());
      bps.add(getSecondLoc());
    } else if (getTransType() == Type.for5toRev3) {
      bps.add(new Location(getChr(), getCoord() - 1));
      bps.add(new Location(getSecondChr(), getSecondCoord() - 1));
    }
    return bps;

  }

  @Override
  public String getSdiString() {
    return null;
  }

  @Override
  public String getVariantIdString() {
    String variantIdString =
        "XLOC\t" + getLoc() + "\t" + getSecondLoc() + "\t" + getTransType() + "\t " + getHeterogeneity();
    if (getBpReads() != null) {
      variantIdString += "\t [";
      for (BreakpointReads bpReads : getBpReads()) {
        variantIdString += bpReads + " ";
      }
      variantIdString += "]";
    }
    variantIdString += "\t" + isIndel();
    return variantIdString;
  }

  @Override
  public boolean equals(Object obj) {
    if (obj == null || !(obj instanceof Translocation)) {
      return false;
    } else {
      Translocation other = (Translocation) obj;
      return getLoc().equals(other.getLoc()) && getHeterogeneity() == other.getHeterogeneity()
          && isIndel() == other.isIndel() && getSecondLoc().equals(other.getSecondLoc())
          && getTransType() == other.getTransType();
    }
  }

  @Override
  public String toString() {
    switch (getTransType()) {
      case for3toFor5:
        return String.format("%-6s%s - %s (forward/3' to forward/5')", "XLOC", getLoc(), getSecondLoc());
      case for3toRev5:
        return String.format("%-6s%s - %s (forward/3' to reverse/5')", "XLOC", getLoc(), getSecondLoc());
      case for5toRev3:
        return String.format("%-6s%s - %s (forward/5' to reverse/3')", "XLOC", getLoc(), getSecondLoc());
      default:
        throw new ShearRuntimeException("Illegal translocation type specified");
    }
  }

}
