/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio.shear.variants;

import java.util.List;

import edu.umn.cs.kumarbio.Chromosome;
import edu.umn.cs.kumarbio.shear.ShearRuntimeException;
import edu.umn.cs.kumarbio.shear.heterogeneity.BreakpointReads;

/**
 * Represents a variant in a genome. This is an abstract class that is extended
 * for particular types of variants.
 */
public abstract class Variant implements Comparable<Variant> {

  // The location of this variant (i.e. first affected base in the chromosome)
  private Location              loc;

  // The heterogeneity level of this variant (-1 indicates it is not set yet)
  private double                heterogeneity;

  // The breakpoint read information associated with this variant
  private List<BreakpointReads> bpReads;

  // Indicates whether or not this variant is an indel (i.e. small insertion or deletion)
  private boolean               isIndel;

  /**
   * Create a new Variant object, initializing the isIndel parameter to false,
   * the heterogeneity level to -1 (represents null), and the breakpoint reads
   * to null.
   *
   * @param loc
   *          The location of this variant (i.e. first affected base in the
   *          chromosome)
   */
  public Variant(Location loc) {
    this(loc, false);
  }

  /**
   * Create a new Variant object, initializing the heterogeneity level to -1
   * (represents null) and the breakpoint reads to null.
   *
   * @param loc
   *          The location of this variant (i.e. first affected base in the
   *          chromosome)
   * @param isIndel
   *          Whether or not this variant is an indel (i.e. small insertion or
   *          deletion)
   */
  public Variant(Location loc, boolean isIndel) {

    // Parse arguments and throw errors
    if (loc == null) {
      throw new IllegalVariantException("Null location specified for variant");
    }

    // Initialize variables
    this.loc = loc;
    heterogeneity = -1d;
    bpReads = null;
    this.isIndel = isIndel;

  }

  /**
   * Sets the heterogeneity level of this variant.
   *
   * @param heterogeneity
   *          The heterogeneity level to set for this variant (must be between 0
   *          and 1, inclusive)
   */
  public void setHeterogeneity(double heterogeneity) {
    if (heterogeneity < 0d || heterogeneity > 1d) {
      throw new IllegalArgumentException("Invalid heterogeneity level set for variant: " + heterogeneity);
    } else {
      this.heterogeneity = heterogeneity;
    }
  }

  /**
   * Sets the breakpoint reads information for this variant.
   *
   * @param bpReads
   *          The breakpoint read information associated with this variant
   */
  public void setBpReads(List<BreakpointReads> bpReads) {
    if (bpReads.size() == 0) {
      throw new IllegalArgumentException("Must add at least one BreakpointReads object");
    } else {
      this.bpReads = bpReads;
    }
  }

  /**
   * Returns the chromosome that this variant occurs on.
   *
   * @return The chromosome that this variant occurs on
   */
  public Chromosome getChr() {
    return getLoc().getChr();
  }

  /**
   * Returns the coordinate that this variant occurs at (i.e. first affected
   * base in the chromosome).
   *
   * @return The coordinate that this variant occurs at (i.e. first affected
   *         base in the chromosome)
   */
  public long getCoord() {
    return getLoc().getCoord();
  }

  /**
   * Returns the location of this variant (i.e. first affected base in the
   * chromosome).
   *
   * @return The location of this variant (i.e. first affected base in the
   *         chromosome)
   */
  public Location getLoc() {
    return loc;
  }

  /**
   * Returns the heterogeneity level of this variant.
   *
   * @return The heterogeneity level of this variant
   */
  public double getHeterogeneity() {
    return heterogeneity;
  }

  /**
   * Returns the breakpoint read information associated with this variant.
   *
   * @return The breakpoint read information associated with this variant
   */
  public List<BreakpointReads> getBpReads() {
    return bpReads;
  }

  /**
   * Returns true if this variant is an indel (i.e. small insertion or
   * deletion), false otherwise.
   *
   * @return True if this variant is an indel (i.e. small insertion or
   *         deletion), false otherwise
   */
  public boolean isIndel() {
    return isIndel;
  }

  /**
   * Returns a string containing a properly-formatted SDI line for a particular
   * variant specified by a location, string of reference bases, and string of
   * consensus (i.e. variant) bases. This method will format the information
   * into SDI format and calculate the variant size difference, which is the
   * third field in an SDI line.
   *
   * @param loc
   *          The location of this variant
   * @param reference
   *          The reference bases at this location (can be "-" to specify no
   *          changed reference bases)
   * @param consensus
   *          The consensus/variant bases that are replacing the reference bases
   *          at this location (can be "-" to specify a deletion of reference
   *          bases)
   * @return A properly-formatted SDI line for the specified variant information
   */
  public String formatSdiLine(Location loc, String reference, String consensus) {

    // StringBuffer to hold SDI line
    StringBuffer line = new StringBuffer();

    // Chromosome name and coordinate
    line.append(loc.getChr().getName());
    line.append("\t");
    line.append(loc.getCoord());
    line.append("\t");

    // Calculate length of variant (i.e. difference between consensus and reference sequences)
    if (reference.equals("-")) {
      if (consensus.equals("-")) {
        line.append(0);
      } else {
        line.append(consensus.length());
      }
    } else if (consensus.equals("-")) {
      if (reference.equals("-")) {
        line.append(0);
      } else {
        line.append("-");
        line.append(reference.length());
      }
    } else {
      line.append(consensus.length() - reference.length());
    }
    line.append("\t");

    // Reference and consensus bases and end of SDI string
    line.append(reference);
    line.append("\t");
    line.append(consensus);
    line.append("\t*\t*");

    // Return SDI line
    return line.toString();

  }

  /**
   * Returns a list of breakpoint locations associated with this variant. A
   * breakpoint location is a coordinate just before the break.
   *
   * @return A list of breakpoint locations associated with this variant
   */
  public abstract List<Location> getBreakpoints();

  /**
   * Returns a string containing the SDI format of this variant.
   *
   * @return A string containing the SDI format of this variant
   */
  public abstract String getSdiString();

  /**
   * This returns a string that contains all of the information about this
   * variant that would make it unique (used in compareTo method).
   *
   * @return The variant ID string for this variant
   */
  public abstract String getVariantIdString();

  @Override
  public int compareTo(Variant other) {

    // Return 0 if variants are equal
    if (equals(other)) {
      return 0;
    }

    // Return whichever has first location
    if (getLoc().compareTo(other.getLoc()) < 0) {
      return -1;
    } else if (getLoc().compareTo(other.getLoc()) > 0) {
      return +1;
    }

    // If exact same location, then return based on variant ID string
    int comparison = getVariantIdString().compareTo(other.getVariantIdString());
    if (comparison == 0) {
      throw new ShearRuntimeException(
          "Unexpected behavior while comparing order of Variant objects (variant ID string comparison returned 0)");
    }
    return comparison;

  }

}
