/*
 * Sean Landman
 * Data Mining for Biomedical Informatics Group
 * University of Minnesota - Twin Cities
 */

package edu.umn.cs.kumarbio.shear.variants;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import edu.umn.cs.kumarbio.shear.heterogeneity.BreakpointReads;

/**
 * Represents a SNP in a genome.
 */
public class SNP extends Variant {

  // The reference base that is being substituted for
  private char refBase;

  // The new base that is being substituted in
  private char varBase;

  /**
   * Creates a new SNP object.
   *
   * @param loc
   *          The location of this SNP (i.e. the base that is substituted)
   * @param varBase
   *          The new base that is being substituted in
   * @throws IOException
   *           If there are problems reading from the FASTA file to determine
   *           the reference base
   */
  public SNP(Location loc, char varBase) throws IOException {
    super(loc);

    // Parse arguments and throw errors
    if (varBase != 'A' && varBase != 'C' && varBase != 'G' && varBase != 'T') {
      throw new IllegalVariantException("Invalid variant based specified for SNP: " + varBase);
    }

    // Initialize variables
    refBase = loc.getChr().getSequence(loc.getCoord(), loc.getCoord()).charAt(0);
    this.varBase = varBase;

  }

  /**
   * Returns the reference base that is being substituted for.
   *
   * @return The reference base that is being substituted for
   */
  public char getRefBase() {
    return refBase;
  }

  /**
   * Returns the new base that is being substituted in
   *
   * @return The new base that is being substituted in
   */
  public char getVarBase() {
    return varBase;
  }

  @Override
  public List<Location> getBreakpoints() {

    // No breakpoints for SNP
    return new ArrayList<Location>();

  }

  @Override
  public String getSdiString() {
    return formatSdiLine(getLoc(), String.valueOf(getRefBase()), String.valueOf(getVarBase()));
  }

  @Override
  public String getVariantIdString() {
    String variantIdString =
        "SNP\t" + getLoc() + "\t" + getRefBase() + "\t" + getVarBase() + "\t " + getHeterogeneity();
    if (getBpReads() != null) {
      variantIdString += "\t [";
      for (BreakpointReads bpReads : getBpReads()) {
        variantIdString += bpReads + " ";
      }
      variantIdString += "]";
    }
    variantIdString += "\t" + isIndel();
    return variantIdString;
  }

  @Override
  public boolean equals(Object obj) {
    if (obj == null || !(obj instanceof SNP)) {
      return false;
    } else {
      SNP other = (SNP) obj;
      return getLoc().equals(other.getLoc()) && getHeterogeneity() == other.getHeterogeneity()
          && isIndel() == other.isIndel() && getRefBase() == other.getRefBase() && getVarBase() == other.getVarBase();
    }
  }

  @Override
  public String toString() {
    return String.format("%-6s%s:%d (%s>%s)", "SNP", getChr().getName(), getCoord(), getRefBase(), getVarBase());
  }

}
